from datetime import timedelta, datetime
from django.db.models.fields import Field
from django.utils.encoding import force_str
from time import sleep
from typing import Type, List, Tuple, Any, Optional, Sequence
from django.db.models import Model
from django.utils.timezone import now
from jutil.format import choices_label


def get_object_or_none(cls: Any, **kwargs) -> Any:
    """
    Returns model instance or None if not found.
    :param cls: Class or queryset
    :param kwargs: Filters for get() call
    :return: Object or None
    """
    try:
        qs = cls._default_manager.all() if hasattr(cls, "_default_manager") else cls  # pylint: disable=protected-access
        return qs.get(**kwargs)
    except Exception:
        return None


def wait_object_or_none(cls: Any, timeout: float = 5.0, sleep_interval: float = 1.0, **kwargs) -> Any:
    """
    Returns model instance or None if not found after specified timeout.
    Waits timeout before returning if no object available.
    Waiting is done by sleeping specified intervals.
    :param cls: Class or queryset
    :param timeout: Timeout in seconds
    :param sleep_interval: Sleep interval in seconds
    :param kwargs: Filters for get() call
    :return: Object or None
    """
    t0: Optional[datetime] = None
    t1: Optional[datetime] = None
    qs0 = cls._default_manager if hasattr(cls, "_default_manager") else cls  # pylint: disable=protected-access
    while t0 is None or t0 < t1:  # type: ignore
        obj = qs0.all().filter(**kwargs).first()
        if obj is not None:
            return obj
        t0 = now()
        if t1 is None:
            t1 = t0 + timedelta(seconds=timeout)
        sleep(sleep_interval)
    return qs0.all().filter(**kwargs).first()


def get_model_field_or_none(instance, field_name: str) -> Optional[Field]:
    """
    Returns model field.
    :param instance: Model instance
    :param field_name: Model attribute name
    :return: Optional[Field]
    """
    for f in instance._meta.fields:
        if f.attname == field_name:
            return f
    return None


def get_model_field(instance, field_name: str) -> Field:
    """
    Returns model field.
    :param instance: Model instance
    :param field_name: Model attribute name
    :return: Field
    """
    f = get_model_field_or_none(instance, field_name)
    if f is None:
        raise Exception(f"Field {field_name} not found from {instance}")
    return f


def get_model_field_label(instance, field_name: str) -> str:
    """
    Returns model field label.
    Resolves also function / property short_description meta fields.
    :param instance: Model instance
    :param field_name: Model attribute name
    :return: label str
    """
    for f in instance._meta.fields:
        if field_name in (f.name, f.attname):
            return f.verbose_name
    try:
        return getattr(getattr(getattr(instance.__class__, field_name), "fget"), "short_description")
    except AttributeError:
        pass
    try:
        return getattr(getattr(instance.__class__, field_name), "short_description")
    except AttributeError:
        pass
    return field_name


def get_model_field_label_and_value(instance, field_name: str) -> Tuple[str, str]:
    """
    Returns model field label and value as str.
    :param instance: Model instance
    :param field_name: Model attribute name
    :return: (label, value) tuple
    """
    label = get_model_field_label(instance, field_name)
    value = str(getattr(instance, field_name)) if hasattr(instance, field_name) else None
    for f in instance._meta.fields:
        if field_name in (f.attname, f.name):
            if hasattr(f, "choices") and f.choices and value is not None:
                value = choices_label(f.choices, value)
            break
    if value is None:
        value = ""
    val = force_str(value)
    return label, val


def is_model_field_changed(instance, field_name: str) -> bool:
    """
    Compares model instance field value to value stored in DB.
    If object has not been stored in DB (yet) field is considered unchanged.
    :param instance: Model instance
    :param field_name: Model attribute name
    :return: True if field value has been changed compared to value stored in DB.
    """
    assert hasattr(instance, field_name)
    if not hasattr(instance, "pk") or instance.pk is None:
        return False
    qs = instance.__class__.objects.all()
    params = {"pk": instance.pk, field_name: getattr(instance, field_name)}
    return not qs.filter(**params).exists()


def get_model_field_names(instance, cls: Optional[Type[Model]] = None, exclude_fields: Sequence[str] = ("id",), base_class_suffix: str = "_ptr") -> List[str]:
    if cls is None:
        cls = instance.__class__
    return [f.name for f in cls._meta.fields if f.name not in exclude_fields and not f.name.endswith(base_class_suffix)]


def clone_model(
    instance, cls: Optional[Type[Model]] = None, commit: bool = True, exclude_fields: Sequence[str] = ("id",), base_class_suffix: str = "_ptr", **kw
):
    """
    Assigns model fields to new object. Ignores exclude_fields list and
    attributes ending with pointer suffix (default '_ptr')
    :param instance: Instance to copy
    :param cls: Class (opt)
    :param commit: Save or not
    :param exclude_fields: List of fields to exclude
    :param base_class_suffix: End of name for base class pointers, e.g. model Car(Vehicle) has implicit vehicle_ptr
    :return: New instance
    """
    if cls is None:
        cls = instance.__class__
    keys = get_model_field_names(instance, cls=cls, exclude_fields=exclude_fields, base_class_suffix=base_class_suffix)
    new_instance = cls()
    for k in keys:
        setattr(new_instance, k, getattr(instance, k))
    for k, v in kw.items():
        setattr(new_instance, k, v)
    if commit:
        new_instance.save()
    return new_instance



import logging
from collections import OrderedDict
from typing import Dict, TypeVar

R = TypeVar("R")
S = TypeVar("S")

logger = logging.getLogger(__name__)


def sorted_dict(d: Dict[S, R]) -> Dict[S, R]:
    """
    Returns dict sorted by ascending key
    :param d: dict
    :return: OrderedDict
    """
    return OrderedDict(sorted(d.items()))



import logging
import traceback
from typing import Dict, Optional
from urllib.parse import urlencode
from django.conf import settings
from django.http import HttpRequest
from django.utils import timezone
from ipware import get_client_ip  # type: ignore
from jutil.email import send_email

logger = logging.getLogger(__name__)


class EnsureOriginMiddleware:
    """
    Ensures that request META 'HTTP_ORIGIN' is set.
    Uses request get_host() to set it if missing.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        if not request.META.get("HTTP_ORIGIN", None):
            request.META["HTTP_ORIGIN"] = request.get_host()

        # get response
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        return response


class LogExceptionMiddleware:
    """
    Logs exception and sends email to admins about it.
    Uses list of emails from settings.ADMINS.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_exception(self, request, e):
        """
        Logs exception error message and sends email to ADMINS if hostname is not testserver and DEBUG=False.
        :param request: HttpRequest
        :param e: Exception
        """
        assert isinstance(request, HttpRequest)
        method = str(request.method).upper()
        uri = request.build_absolute_uri()
        user = request.user
        msg = "{method} {uri}\n{err} (IP={ip}, user={user}) {trace}".format(
            method=method, uri=uri, user=user, ip=get_client_ip(request)[0], err=e, trace=traceback.format_exc()
        )
        logger.error(msg)
        hostname = request.get_host()
        if not settings.DEBUG and hostname != "testserver":
            send_email(settings.ADMINS, "Error @ {}".format(hostname), msg)


class EnsureLanguageCookieMiddleware:
    """
    Ensures language cookie (by name settings.LANGUAGE_COOKIE_NAME) is set.
    Sets it as settings.LANGUAGE_CODE if missing.
    Allows changing settings by passing querystring parameter named settings.LANGUAGE_COOKIE_NAME
    (default: django_language).

    Order of preference for the language (must be one of settings.LANGUAGES to be used):
    1) Explicit querystring GET parameter (e.g. ?lang=en)
    2) Previously stored cookie
    3) settings.LANGUAGE_CODE
    """

    _languages: Optional[Dict[str, str]]

    def __init__(self, get_response=None):
        self.get_response = get_response
        self._languages = None

    @property
    def languages(self) -> Dict[str, str]:
        if self._languages is None:
            self._languages = dict(settings.LANGUAGES)
        return self._languages

    def __call__(self, request):
        lang_cookie_name = settings.LANGUAGE_COOKIE_NAME if hasattr(settings, "LANGUAGE_COOKIE_NAME") else "django_language"
        lang_cookie = request.COOKIES.get(lang_cookie_name)
        lang = request.GET.get(lang_cookie_name)
        if not lang:
            lang = lang_cookie
        if not lang or lang not in self.languages:
            lang = settings.LANGUAGE_CODE if hasattr(settings, "LANGUAGE_CODE") else "en"
        request.COOKIES[lang_cookie_name] = lang

        res = self.get_response(request)

        if lang_cookie is None or lang != lang_cookie:
            secure = hasattr(settings, "LANGUAGE_COOKIE_SECURE") and settings.LANGUAGE_COOKIE_SECURE
            httponly = hasattr(settings, "LANGUAGE_COOKIE_HTTPONLY") and settings.LANGUAGE_COOKIE_HTTPONLY
            res.set_cookie(lang_cookie_name, lang, secure=secure, httponly=httponly)
        return res


class ActivateUserProfileTimezoneMiddleware:
    """
    Uses 'timezone' string in request.user.profile to activate user-specific timezone.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        activated = False
        if request.user.is_authenticated:
            user = request.user
            if hasattr(user, "profile") and user.profile:
                up = user.profile
                if hasattr(up, "timezone") and up.timezone:
                    timezone.activate(up.timezone)
                    activated = True
                else:
                    logger.warning("User profile timezone missing so user.profile.timezone could not be activated")
            else:
                logger.warning("User profile missing so user.profile.timezone could not be activated")

        # get response
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        if activated:
            timezone.deactivate()
        return response


class TestClientLoggerMiddleware:
    """
    Logs requests to be used with Django test framework client.
    """

    ignored_paths = {
        "/admin/jsi18n/",
        "/favicon.ico",
    }

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        if settings.DEBUG:
            assert isinstance(request, HttpRequest)
            if request.path not in self.ignored_paths:
                url = request.path
                qs = request.GET.dict()
                if qs:
                    url += "?" + urlencode(request.GET.dict())
                logger.debug('self.client.%s("%s", data=%s)', request.method.lower(), url, request.POST.dict())
        response = self.get_response(request)
        return response
